package software.amazon.smithy.go.codegen;

import software.amazon.smithy.codegen.core.directed.GenerateServiceDirective;

public class ShapeDirectedCodegen extends AbstractDirectedCodegen {
    @Override
    public void generateService(GenerateServiceDirective<GoCodegenContext, GoSettings> directive) {
        // pass
    }
}
